@@ -5,7 +5,7 @@
 
 /*****************************************************************************/
 /*  bt_for_ap.h - branch trace start/stop function for application           */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2006                         */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2007                         */
 /*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
@@ -81,7 +81,7 @@ static int bt_format_cmd_and_exec(char *fmt, ...)
 
 static int bt_write_enable(int enable)
 {
-	int fd, rc;
+	int fd;
 	char *s;
 
 	fd = open("/proc/btrax/enable", O_WRONLY);
@@ -93,13 +93,11 @@ static int bt_write_enable(int enable)
 	return 0;
 }
 
-static inline void __bt_start_from_ap(int n, va_list argp)
+static inline void __bt_start_from_ap(pid_t pid, char *syscall_ids)
 {
 	char *outdir, out[BT_MAX_BUF_LEN];
 	struct stat st;
-	pid_t pid;
-	int i, len, left, rc;
-	char *p;
+	int rc;
 
 	if (stat("/proc/btrax/enable", &st) < 0) {
 		/*
@@ -108,46 +106,29 @@ static inline void __bt_start_from_ap(int n, va_list argp)
 			*/
 		return;
 	}
-	pid = getpid();
 	if (!(outdir = getenv("BT_OUTPUT_DIR"))) {
 		dprintf(stderr, "BTRAX: can't get 'BT_OUTPUT_DIR' value\n");
 		return;
 	}
-	if (snprintf(out, BT_MAX_BUF_LEN, "%s/%d.maps", outdir, pid)
-	    == BT_MAX_BUF_LEN) {
-		dprintf(stderr, "BTRAX: BT_MAX_BUF_LEN is too short\n");
-		return;
-	}
-	if (stat(out, &st) < 0) {
-		if (bt_format_cmd_and_exec("cp /proc/%d/maps %s", pid, out) < 0)
+	if (!syscall_ids) {
+		if (snprintf(out, BT_MAX_BUF_LEN, "%s/%d.maps", outdir, pid)
+		    == BT_MAX_BUF_LEN) {
+			dprintf(stderr, "BTRAX: BT_MAX_BUF_LEN is too short\n");
 			return;
+		}
+		if (stat(out, &st) < 0) {
+			if (bt_format_cmd_and_exec("cat /proc/%d/maps > %s",
+						   pid, out) < 0)
+				return;
+		}
 	}
 	if (bt_format_cmd_and_exec("echo %d > /proc/btrax/pid", pid) < 0)
 		return;
-	if (n > 0) {
-		p = out;
-		left = BT_MAX_BUF_LEN;
-
-		rc = snprintf(p, left, "echo ");
-		p += rc;
-		left -= rc;
-
-		for (i = 0; i < n; i++) {
-			rc = snprintf(p, left, "%d,", va_arg(argp, int));
-			if (rc >= left) {
-				dprintf(stderr,
-					"BTRAX: BT_MAX_BUF_LEN is too short\n");
-				return;
-			}
-			p += rc;
-			left -= rc;
-		}
-		/* delete last ',' character */
-		p--;
-		left++;
-
-		rc = snprintf(p, left, " > /proc/btrax/filter_syscall");
-		if (rc >= left) {
+	if (syscall_ids) {
+		rc = snprintf(out, BT_MAX_BUF_LEN,
+			      "echo %s > /proc/btrax/filter_syscall",
+			      syscall_ids);
+		if (rc >= BT_MAX_BUF_LEN) {
 			dprintf(stderr, "BTRAX: BT_MAX_BUF_LEN is too short\n");
 			return;
 		}
@@ -158,7 +139,7 @@ static inline void __bt_start_from_ap(int n, va_list argp)
 
 void bt_start_from_ap(void)
 {
-	__bt_start_from_ap(0, NULL);
+	__bt_start_from_ap(getpid(), NULL);
 }
 
 void bt_stop_from_ap(void)
@@ -166,20 +147,39 @@ void bt_stop_from_ap(void)
 	bt_write_enable(0);
 }
 
+#if 0
 void bt_start_syscall_chk(int n, ...)
 {
 	va_list argp;
+	char *p, tmp[BT_MAX_BUF_LEN];
+	int i, rc, left;
+
+	p = tmp;
+	left = BT_MAX_BUF_LEN;
 
 	va_start(argp, n);
-	__bt_start_from_ap(n, argp);
+	for (i = 0; i < n; i++) {
+		rc = snprintf(p, left, "%d,", va_arg(argp, int));
+		if (rc >= left) {
+			dprintf(stderr,
+				"BTRAX: BT_MAX_BUF_LEN is too short\n");
+			return;
+		}
+		p += rc;
+		left -= rc;
+	}
+	/* delete last ',' character */
+	p--;
+	*p = '\0';
 	va_end(argp);
+
+	__bt_start_from_ap(getpid(), tmp);
 }
+#endif
 
 #define BT_COLLECT_NAME	"bt_collect_log"
 void bt_term_from_ap(void)
 {
-	struct stat st;
-
 	if (bt_write_enable(0) < 0) {
 		if (seteuid(getuid()) < 0)
 			return;
