@@ -5,7 +5,7 @@
 
 /*****************************************************************************/
 /*  bt_on_off.c - branch trace module on-off driver                          */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2006                         */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2007                         */
 /*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
@@ -24,9 +24,8 @@
 /*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA      */
 /*****************************************************************************/
 
-#include <linux/version.h>
 #include <linux/module.h>
-#ifdef USE_KPROBES
+#ifdef USE_SYS_KPROBES
 #  include <linux/kprobes.h>
 #else
 #  include "djprobe.h"
@@ -85,7 +84,7 @@ static void bt_off_probe(struct pt_regs *regs)
 	preempt_enable();
 }
 
-#ifdef USE_KPROBES
+#ifdef USE_SYS_KPROBES
 static struct kprobe on_kp, off_kp;
 static int bt_on_kp_wrapper(struct kprobe *kp, struct pt_regs *regs)
 {
@@ -116,7 +115,7 @@ static int bt_on_off_init(void)
 	if (on_addr) {
 		if (!on_size)
 			return -EINVAL;
-#ifdef USE_KPROBES
+#ifdef USE_SYS_KPROBES
 		on_kp.addr = (kprobe_opcode_t*)on_addr;
 		on_kp.pre_handler = bt_on_kp_wrapper;
 		if ((rc = register_kprobe(&on_kp)) < 0)
@@ -133,7 +132,7 @@ static int bt_on_off_init(void)
 	if (off_addr) {
 		if (!off_size)
 			return -EINVAL;
-#ifdef USE_KPROBES
+#ifdef USE_SYS_KPROBES
 		off_kp.addr = (kprobe_opcode_t*)off_addr;
 		off_kp.pre_handler = bt_off_kp_wrapper;
 		if ((rc = register_kprobe(&off_kp)) < 0)
@@ -153,7 +152,7 @@ static int bt_on_off_init(void)
 static void bt_on_off_cleanup(void)
 {
 	if (on_addr && on_size) {
-#ifdef USE_KPROBES
+#ifdef USE_SYS_KPROBES
 		unregister_kprobe(&on_kp);
 #else
 		unregister_djprobe(&on_djp);
@@ -161,7 +160,7 @@ static void bt_on_off_cleanup(void)
 		dprintk("bt_on_probe unregistered.\n");
 	}
 	if (off_addr && off_size) {
-#ifdef USE_KPROBES
+#ifdef USE_SYS_KPROBES
 		unregister_kprobe(&off_kp);
 #else
 		unregister_djprobe(&off_djp);
