@@ -5,7 +5,7 @@
 
 /*****************************************************************************/
 /*  bt_coverage.c - coverage information display program                     */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2006                         */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2007                         */
 /*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
@@ -30,7 +30,7 @@
 #include "bt_hconv.h"
 
 #define BT_COVERAGE_VER	BT_VER
-#define	COPYRIGHT	"Copyright (c) Hitachi, Ltd. 2005-2006"
+#define	COPYRIGHT	"Copyright (c) Hitachi, Ltd. 2005-2007"
 
 int output_summary = 0;
 int verbose = 0;
@@ -38,8 +38,9 @@ char *outdir;
 char *srcdir;
 char *includes;
 char *excludes;
-node *include_funcs;
+node *include_funcs;	/* avl tree of unsigned long address */
 node *exclude_funcs;
+node *include_fcs;	/* avl tree of func_chk */
 struct r2n_info r2n_info = { NULL, 128, 0 };
 #define HTML_OUTPUT	(outdir != NULL)
 #define LIMIT_BY_FUNCS	(include_funcs != NULL)
@@ -90,10 +91,10 @@ int proc_each_record(struct bt_record *__rec, off_t i_rec, void *tmp)
 
 	if (is_warn_record(&rec)) {
 		struct warn_record *warn = (struct warn_record*)&rec;
-		fprintf(stderr, "WARN: bts left only: %ld\n", warn->left);
+		printf("WARN: bts left only: %ld\n", warn->left);
 		goto EXIT;
 	}
-	if (is_pid_record(&rec))
+	if (is_pid_record(&rec) || is_comm_record(&rec))
 		goto EXIT;
 	rc = chk_fix_from_cache(&r2n_info, &rec.from, rec.to,
 				&r2n, &p_from, &idx_from);
@@ -143,11 +144,9 @@ int proc_each_record(struct bt_record *__rec, off_t i_rec, void *tmp)
 				 * following processing is done.
 				 */
 				if (strcmp(r2n->basename, MOD_NAME ".ko") != 0){
-					fprintf(stderr,
-						"WARN: detect lack of log");
-					fprintf(stderr,
-						" (0x%08lx) -> 0x%08lx -> 0x%08lx\n",
-						data->last, rec.from, rec.to);
+					printf("WARN: detect lack of log");
+					printf(" (0x%08lx) -> 0x%08lx -> 0x%08lx\n",
+					       data->last, rec.from, rec.to);
 					break;
 					/*
 					fprintf(stdout, "!!! NG path tree\n");
@@ -203,7 +202,7 @@ void printf_funcname(struct bfd_if *bi, bfd_vma addr)
 	if (rc >= 0) {
 		printf_func_name(func_name, offset);
 		rc = get_source_info(bi, addr, &src_name, &func_name, &lno);
-		if (rc >= 0 && src_name)
+		if (rc >= 0 && src_name && lno)
 			printf(":%s,%d", src_name, lno);
 	} else
 		printf_bfd_vma(addr);
@@ -215,7 +214,7 @@ void printf_srcname_and_lno(struct bfd_if *bi, bfd_vma addr)
 	const char *src_name, *func_name;
 
 	rc = get_source_info(bi, addr, &src_name, &func_name, &lno);
-	if (rc >= 0 && src_name)
+	if (rc >= 0 && src_name && lno)
 		printf("%s,%d", src_name, lno);
 	else
 		printf_bfd_vma(addr);
@@ -259,7 +258,7 @@ int create_filter_funcs(char *funcs, node **list)
 		p = strtok(NULL, ",");
 	}
 	if (!*list) {
-		fprintf(stderr, "function list is empty.\n");
+		printf("WARN: include functions were not found.\n");
 		return -1;
 	}
 	return 0;
@@ -280,13 +279,6 @@ int is_exclude_faddr(unsigned long addr)
  *  display function coverage
  *-----------------------------------------------------------------------------
  */
-struct func_chk {
-	unsigned long	addr;
-	unsigned long	cnt;
-	int		invalid;
-	node		*childs;
-};
-
 static int f_cmp_addr2fc(void *__target, void *__dt)
 {
 	unsigned long addr = (unsigned long)__target;
@@ -323,9 +315,11 @@ int __check_func(node **fc_list, unsigned long addr, unsigned long cnt)
 	}
 	fc = xmalloc(sizeof(struct func_chk));
 	fc->addr = addr;
+	fc->end = 0;
 	fc->cnt = cnt;
-	fc->invalid = 0;
+	SET_FC_VALID(fc);
 	fc->childs = NULL;
+	fc->excluded_tree_cnt = 0;
 
 	*fc_list = insert_tree(fc, *fc_list, f_cmp_fc);
 	if (!(*fc_list))
@@ -363,10 +357,68 @@ int check_func(node **fc_list, struct path *p)
 	return 0;
 }
 
-int mark_valid_path(struct range_to_name *r2n, struct func_chk *fc,
-		    unsigned long end, node **fc_list);
+struct get_valid_funcs_pack {
+	node		*valid_funcs;
+	func_compare	f_cmp;
+};
 
-int __mark_valid_path(struct func_chk *fc, unsigned long faddr, node **fc_list)
+static void f_free_unnecessary_fc(void *__dt)
+{
+	struct func_chk *fc = (struct func_chk*)__dt;
+
+	if (!IS_CHECKED_FC(fc))
+		free(fc);
+}
+
+static int f_slim_down_flist(void *__dt, void *user_data)
+{
+	struct func_chk *fc = (struct func_chk*)__dt;
+	struct get_valid_funcs_pack *pack =
+		(struct get_valid_funcs_pack*)user_data;
+
+	if (!IS_CHECKED_FC(fc))
+		return 0;
+	pack->valid_funcs = insert_tree(fc, pack->valid_funcs, pack->f_cmp);
+	return 0;
+}
+
+static void slim_down_flist(node **fc_list, func_compare f_compare)
+{
+	struct get_valid_funcs_pack pack;
+
+	pack.valid_funcs = NULL;
+	pack.f_cmp = f_compare;
+	for_each_node(*fc_list, f_slim_down_flist, &pack);
+	free_tree(*fc_list, f_free_unnecessary_fc);
+	*fc_list = pack.valid_funcs;
+}
+
+static int f_cmp_fc_excluded_tcnt(void *__target, void *__dt)
+{
+	struct func_chk *t_fc = (struct func_chk*)__target;
+	struct func_chk *fc = (struct func_chk*)__dt;
+
+	if (t_fc->excluded_tree_cnt > fc->excluded_tree_cnt)
+		return -1;
+	if (t_fc->excluded_tree_cnt < fc->excluded_tree_cnt)
+		return 1;
+	if (t_fc->addr < fc->addr)
+		return -1;
+	if (t_fc->addr > fc->addr)
+		return 1;
+	return 0;
+}
+
+static inline void get_excluded_tcnt_func_list(node **fc_list)
+{
+	slim_down_flist(fc_list, f_cmp_fc_excluded_tcnt);
+}
+
+int mark_valid_path(struct range_to_name *r2n, struct func_chk *fc,
+		    unsigned long end, node *fc_list);
+
+int __mark_valid_path(struct range_to_name *r2n, struct func_chk *fc,
+		      unsigned long faddr, node *fc_list, int from_uk)
 {
 	struct range_to_name *c_r2n;
 	unsigned long c_end;
@@ -374,24 +426,30 @@ int __mark_valid_path(struct func_chk *fc, unsigned long faddr, node **fc_list)
 
 	if (!is_exclude_faddr(faddr) &&
 	    get_func_info(&r2n_info, faddr, &c_end, &c_r2n)) {
-		c_fc = search_tree((void*)faddr, *fc_list, f_cmp_addr2fc);
+		c_fc = search_tree((void*)faddr, fc_list, f_cmp_addr2fc);
 		if (c_fc) {
+			if (from_uk) {
+				if (IS_CHECKED_FC(c_fc))
+					return 0;
+				if (c_fc->cnt)
+					SET_FC_UT(c_fc);
+				return 0;
+			}
 			fc->childs = insert_tree(c_fc, fc->childs, f_cmp_fc);
 			if (!(fc->childs))
 				return -1;
-			if (c_fc->invalid) {
-				c_fc->invalid = 0;
-				if (mark_valid_path(c_r2n, c_fc, c_end, fc_list)
-				    < 0)
-					return -1;
-			}
+			if (IS_CHECKED_FC(c_fc))
+				return 0;
+			SET_FC_VALID(c_fc);
+			if (mark_valid_path(c_r2n, c_fc, c_end, fc_list) < 0)
+				return -1;
 		}
 	}
 	return 0;
 }
 
 int mark_valid_path(struct range_to_name *r2n, struct func_chk *fc,
-		    unsigned long end, node **fc_list)
+		    unsigned long end, node *fc_list)
 {
 	struct path *p;
 	long i, top;
@@ -405,6 +463,8 @@ int mark_valid_path(struct range_to_name *r2n, struct func_chk *fc,
 			r2n->basename, fc->addr, end);
 		return -1;
 	}
+	fc->i = top;
+	fc->end = end;
 	for (i = top; i < r2n->cnt; i++) {
 		p = r2n->pt[i];
 		if (p->next > end)
@@ -414,15 +474,19 @@ int mark_valid_path(struct range_to_name *r2n, struct func_chk *fc,
 		case BTYPE_CALL:
 		case BTYPE_JMP:
 		case BTYPE_BRANCH:
+			/* Function calls by the indirect addressing are not
+			 * included in the function call tree.
+			 */
 			if (p->jmp == UNKNOWN_BADDR) {
 				for (uk = (struct unknown*)p->jmp_cnt; uk;
 				     uk = uk->next) {
-					if (__mark_valid_path(fc, uk->addr,
-							      fc_list) < 0)
+					if (__mark_valid_path(r2n, fc, uk->addr,
+							      fc_list, 1) < 0)
 						return -1;
 				}
 			} else {
-				if (__mark_valid_path(fc, p->jmp, fc_list) < 0)
+				if (__mark_valid_path(r2n, fc, p->jmp, fc_list,
+						      0) < 0)
 					return -1;
 			}
 			break;
@@ -468,20 +532,23 @@ static void f_fc_free(void *__dt)
 static int f_mark_include_funcs(void *__dt, void *user_data)
 {
 	unsigned long addr = (unsigned long)__dt;
-	node **fc_list = (node**)user_data;
-	struct range_to_name *c_r2n;
-	unsigned long c_end;
+	node *fc_list = (node*)user_data;
+	struct range_to_name *r2n;
+	unsigned long end;
 	struct func_chk *fc;
 	
 	if (is_exclude_faddr(addr) ||
-	    !get_func_info(&r2n_info, addr, &c_end, &c_r2n))
+	    !get_func_info(&r2n_info, addr, &end, &r2n))
 		return 0;
-	fc = search_tree((void*)addr, *fc_list, f_cmp_addr2fc);
-	if (fc && fc->invalid) {
-		fc->invalid = 0;
-		if (mark_valid_path(c_r2n, fc, c_end, fc_list) < 0)
-			return -1;
-	}
+	fc = search_tree((void*)addr, fc_list, f_cmp_addr2fc);
+	if (!fc)
+		return 0;
+	include_fcs = insert_tree(fc, include_fcs, f_cmp_fc);
+	if (IS_CHECKED_FC(fc))
+		return 0;
+	SET_FC_VALID(fc);
+	if (mark_valid_path(r2n, fc, end, fc_list) < 0)
+		return -1;
 	return 0;
 }
 
@@ -489,7 +556,7 @@ static int f_all_fc_mark_invalid(void *__dt, void *user_data)
 {
 	struct func_chk *cc = (struct func_chk*)__dt;
 
-	cc->invalid = 1;
+	SET_FC_NOT_CHECKED(cc);
 	return 0;
 }
 
@@ -542,7 +609,8 @@ int chk_func_coverage(node **fc_list)
 				}
 			}
 		}
-		for_each_node(include_funcs, f_mark_include_funcs, fc_list);
+		for_each_node(include_funcs, f_mark_include_funcs, *fc_list);
+		slim_down_flist(fc_list, f_cmp_fc);
 	}
 	return 0;
 
@@ -564,7 +632,7 @@ static int f_cnt_each_fc(void *__dt, void *user_data)
 	struct func_chk *cc = (struct func_chk*)__dt;
 
 	if (cc->addr >= r2n->begin && cc->addr <= r2n->end) {
-		if (!cc->invalid) {
+		if (IS_VALID_FC(cc)) {
 			pack->all++;
 			if (cc->cnt)
 				pack->called++;
@@ -579,6 +647,23 @@ struct dump_each_fc_data {
 	struct html_out_data	*html_out_dt;
 };
 
+#define __print_one_func_cov(r2n, fc) \
+do { \
+	printf_funcname(&(r2n)->bi, (fc)->addr - (r2n)->offset); \
+	if ((fc)->excluded_tree_cnt > 0) { \
+		if ((fc)->cnt) \
+			printf("\t(%ld, F:%ld)\n", (fc)->cnt, \
+			       (fc)->excluded_tree_cnt); \
+		else \
+			printf("\t(0, F:%ld)\n", (fc)->excluded_tree_cnt); \
+	} else { \
+		if ((fc)->cnt) \
+			printf("\t(%ld)\n", (fc)->cnt); \
+		else \
+			printf("\n"); \
+	} \
+} while(0)
+
 static int f_dump_each_fc(void *__dt, void *user_data)
 {
 	struct dump_each_fc_data *fc_dt = (struct dump_each_fc_data*)user_data;
@@ -586,111 +671,161 @@ static int f_dump_each_fc(void *__dt, void *user_data)
 	struct range_to_name *r2n = fc_dt->r2n;
 
 	if (cc->addr >= r2n->begin && cc->addr <= r2n->end) {
-		if (cc->invalid && !cc->cnt)
+		if (!IS_VALID_FC(cc))
 			return 0;
 		if (HTML_OUTPUT)
-			out_func_html_each(fc_dt->html_out_dt, r2n,
-					   cc->addr - r2n->offset, cc->cnt);
+			out_func_html_each(fc_dt->html_out_dt, r2n, cc);
 		else {
 			printf(cc->cnt ? "(OK) " : "(NT) ");
-			printf_funcname(&r2n->bi, cc->addr - r2n->offset);
-			if (cc->cnt)
-				printf("\t(%ld)\n", cc->cnt);
-			else
-				printf("\n");
+			__print_one_func_cov(r2n, cc);
 		}
 	} else if (cc->addr > r2n->end)
 		return 1;	/* break */
 	return 0;
 }
 
-static void __f_print_func_tree(struct func_chk *fc, int nest)
+struct excluded_cnt_pack {
+	unsigned long		cnt;
+	node			*checked_fc;
+	node			*fc_list;
+};
+
+static int f_cnt_excluded_fc(void *__dt, void *user_data)
 {
-	int i;
+	struct func_chk *fc = (struct func_chk*)__dt;
+	struct excluded_cnt_pack *pack = (struct excluded_cnt_pack*)user_data;
+
+	if (search_tree(fc, pack->checked_fc, f_cmp_fc))
+		return 0;
+	pack->cnt++;
+	pack->checked_fc = insert_tree(fc, pack->checked_fc, f_cmp_fc);
+	return for_each_node(fc->childs, f_cnt_excluded_fc, pack);
+}
+
+static void get_excluded_tree_cnt(node *fc_list, struct func_chk *fc,
+				  unsigned long all_cnt)
+{
+	struct excluded_cnt_pack pack;
+
+	if (fc->excluded_tree_cnt > 0)
+		return;
+	pack.cnt = 0;
+	pack.checked_fc = insert_tree(fc, NULL, f_cmp_fc);
+	pack.fc_list = fc_list;
+	for_each_node(include_fcs, f_cnt_excluded_fc, &pack);
+	free_tree(pack.checked_fc, NULL);
+	fc->excluded_tree_cnt = all_cnt - pack.cnt;
+	return;
+}
+
+struct print_ftree_data {
+	node			*fc_list;
+	node			**printed;
+	int			nest;
+	struct html_out_data	*html_out_dt;
+	unsigned long		tree_cnt;
+};
+
+static void __f_print_func_tree(struct func_chk *fc,
+				struct print_ftree_data *p_ftree_dt,
+				int type, int has_child)
+{
+	int i, nest = p_ftree_dt->nest;
+	struct html_out_data *dt = p_ftree_dt->html_out_dt;
+	node *fc_list = p_ftree_dt->fc_list;
 	struct range_to_name *r2n;
 
 	r2n = addr_to_r2n(&r2n_info, fc->addr);
 	if (!r2n)
 		return;
-	for (i = 0; i < nest; i++)
-		printf("+-");
-	/* COMMON (from) */
-	printf_funcname(&r2n->bi, fc->addr - r2n->offset);
-	if (fc->cnt)
-		printf("\t(%ld)\n", fc->cnt);
-	else
-		printf("\n");
-	/* COMMON (to) */
+	get_excluded_tree_cnt(fc_list, fc, p_ftree_dt->tree_cnt);
+	if (HTML_OUTPUT)
+		out_func_tree_html_each_enter(dt, r2n, fc, nest, type,
+					      has_child);
+	else {
+		switch (type) {
+		case BCOV_TYPE_OK:
+			printf("(OK) ");
+			break;
+		case BCOV_TYPE_NT:
+			printf("(NT) ");
+			break;
+		default:
+			printf("(--) ");
+			break;
+		}
+		for (i = 0; i < nest; i++)
+			printf("+-");
+		__print_one_func_cov(r2n, fc);
+	}
 }
 
-#if 0
 static int f_print_invalid_func(void *__dt, void *user_data)
 {
+	struct print_ftree_data *dt = (struct print_ftree_data*)user_data;
 	struct func_chk *fc = (struct func_chk*)__dt;
+	struct range_to_name *r2n;
 
-	if (fc->invalid && fc->cnt) {
-		printf("(EX) ");
-		__f_print_func_tree(fc, 0);
+	r2n = addr_to_r2n(&r2n_info, fc->addr);
+	if (!r2n)
+		return 0;
+	if (IS_UT_FC(fc)) {
+		if (HTML_OUTPUT)
+			out_func_tree_html_each_invalid(dt->html_out_dt, r2n,
+							fc);
+		else {
+			printf("(UT) ");
+			__print_one_func_cov(r2n, fc);
+		}
 	}
 	return 0;
 }
-#endif
-
-struct print_ftree_data {
-	node	*fc_list;
-	node	**printed;
-	int	nest;
-};
 
 static int f_print_func_tree(void *__dt, void *user_data)
 {
 	struct print_ftree_data dt = *(struct print_ftree_data*)user_data;
 	struct func_chk *fc = (struct func_chk*)__dt;
-	int printed;
+	int printed, type, has_child;
 
 	printed = search_tree((void*)fc->addr, *dt.printed, f_cmp_addr2fc)
 									!= NULL;
 	if (printed) {
 		if (dt.nest) {
-			// Print function name only and continue, cause
-			// to display all child functions.
-			// This output should be delete?
-			printf("(--) ");
-			__f_print_func_tree(fc, dt.nest);
+			has_child = 0;
+			__f_print_func_tree(fc, &dt, BCOV_TYPE_HT, has_child);
+			goto EXIT;
 		}
 		return 0;
 	}
 	*dt.printed = insert_tree(fc, *dt.printed, f_cmp_fc);
 	if (!(*dt.printed))
 		return -1;
-	printf(fc->cnt ? "(OK) " : "(NT) ");
-	__f_print_func_tree(fc, dt.nest);
+	type = fc->cnt ? BCOV_TYPE_OK : BCOV_TYPE_NT;
+	has_child = fc->childs != NULL;
+	__f_print_func_tree(fc, &dt, type, has_child);
 	dt.nest++;
-	return for_each_node(fc->childs, f_print_func_tree, &dt);
+	if (for_each_node(fc->childs, f_print_func_tree, &dt) < 0)
+		return -1;
+	dt.nest--;
+EXIT:
+	if (HTML_OUTPUT)
+		out_func_tree_html_each_exit(dt.html_out_dt, dt.nest,has_child);
+	return 0;
 }
 
-static int print_func_tree_sub(void *__dt, void *user_data)
-{
-	unsigned long addr = (unsigned long)__dt;
-	struct print_ftree_data *dt = (struct print_ftree_data*)user_data;
-	struct func_chk *fc;
-
-	if (is_exclude_faddr(addr))
-		return 0;
-	fc = search_tree((void*)addr, dt->fc_list, f_cmp_addr2fc);
-	if (!fc)
-		return 0;
-	return f_print_func_tree(fc, dt);
-}
-
-void print_func_tree(node *fc_list)
+void print_func_tree(node *fc_list, struct html_out_data *html_out_dt)
 {
 	int i;
 	struct range_to_name *r2n;
 	struct call_cnt_pack pack = { 0, 0, NULL };
 	node *printed = NULL;
 	struct print_ftree_data dt;
+	char *s_inc, *s_exc;
 
+	s_inc = includes ? includes : "(--)";
+	s_exc = excludes ? excludes : "(--)";
+	if (HTML_OUTPUT)
+		out_func_tree_html_start(html_out_dt, s_inc, s_exc);
 	for (i = 0; i < r2n_info.r2n_max; i++) {
 		r2n = r2n_info.all_r2n[i];
 		if (r2n) {
@@ -701,24 +836,25 @@ void print_func_tree(node *fc_list)
 	dt.fc_list = fc_list;
 	dt.printed = &printed;
 	dt.nest = 0;
-	printf("====== includes: %s ======\n", includes);
-	printf("====== excludes: %s ======\n", excludes);
-	printf("====== function tree (%ld/%ld) ======\n",
-	       pack.called, pack.all);
-	for_each_node(include_funcs, print_func_tree_sub, &dt);
-#if 0
-	for_each_node(fc_list, f_print_invalid_func, NULL);
-#endif
+	dt.html_out_dt = html_out_dt;
+	dt.tree_cnt = pack.all;
+	if (!HTML_OUTPUT) {
+		printf("====== includes: %s ======\n", s_inc);
+		printf("====== excludes: %s ======\n", s_exc);
+		printf("====== function tree (%ld/%ld=%.2f%%) ======\n",
+		       pack.called, pack.all,
+		       get_percent(pack.called, pack.all));
+	}
+	for_each_node(include_fcs, f_print_func_tree, &dt);
+	for_each_node(fc_list, f_print_invalid_func, &dt);
+	if (HTML_OUTPUT)
+		out_func_tree_html_end(html_out_dt);
 }
 
 /*-----------------------------------------------------------------------------
  *  display branch coverage
  *-----------------------------------------------------------------------------
  */
-#define IS_SWITCH_JMP(p)	\
-	((p)->type == BTYPE_JMP \
-	 && (p)->jmp == UNKNOWN_BADDR && (p)->next_cnt < 0)
-
 static long get_unknown_bcnt(struct path *p)
 {
 	long i = 0;
@@ -740,9 +876,10 @@ void __dump_one_branch_coverage(struct range_to_name *r2n,
 	else {
 		switch (bcov_type) {
 		case BCOV_TYPE_OK:	printf("(OK) "); break;
-		case BCOV_TYPE_UK:	printf("(UK) "); break;
 		case BCOV_TYPE_HT:	printf("(HT) "); break;
 		case BCOV_TYPE_NT:	printf("(NT) "); break;
+		case BCOV_TYPE_UT:	printf("(UT) "); break;
+		case BCOV_TYPE_UN:	printf("(UN) "); break;
 		}
 		printf_srcname_and_lno(&r2n->bi, base);
 		printf(" [%ld/", b_cnt);
@@ -765,8 +902,9 @@ void __dump_one_branch_coverage(struct range_to_name *r2n,
 	}
 }
 
-void dump_one_branch_coverage(struct range_to_name *r2n,
-			      struct html_out_data *dt, struct path *p)
+inline void do_one_branch_coverage(struct range_to_name *r2n,
+				   struct html_out_data *dt, struct path *p,
+				   func_do_one_bcov func)
 {
 	struct unknown *uk;
 	int bcov_type;
@@ -781,19 +919,18 @@ void dump_one_branch_coverage(struct range_to_name *r2n,
 		if (get_unknown_bcnt(p)) {
 			for (uk = (struct unknown*)p->jmp_cnt; uk;
 			     uk = uk->next) {
-				bcov_type = BCOV_TYPE_UK;
+				bcov_type = BCOV_TYPE_UT;
 				branch = uk->addr - r2n->offset;
 				b_cnt = uk->cnt;
-				__dump_one_branch_coverage(r2n, dt, bcov_type,
-							   base, branch, b_cnt,
-							   fall, f_cnt);
+				func(r2n, dt, bcov_type, base, branch, b_cnt,
+				     fall, f_cnt);
 			}
 		} else {
-			bcov_type = BCOV_TYPE_NT;
+			bcov_type = BCOV_TYPE_UN;
 			branch = UNKNOWN_BADDR;
 			b_cnt = 0;
-			__dump_one_branch_coverage(r2n, dt, bcov_type, base,
-						   branch, b_cnt, fall, f_cnt);
+			func(r2n, dt, bcov_type, base, branch, b_cnt,
+			     fall, f_cnt);
 		}
 	} else {
 		if (p->jmp_cnt != 0 && p->next_cnt != 0) {
@@ -807,11 +944,16 @@ void dump_one_branch_coverage(struct range_to_name *r2n,
 		b_cnt = p->jmp_cnt;
 		fall = p->next - r2n->offset;
 		f_cnt = p->next_cnt;
-		__dump_one_branch_coverage(r2n, dt, bcov_type, base,
-					   branch, b_cnt, fall, f_cnt);
+		func(r2n, dt, bcov_type, base, branch, b_cnt, fall, f_cnt);
 	}
 }
 
+void dump_one_branch_coverage(struct range_to_name *r2n,
+			      struct html_out_data *dt, struct path *p)
+{
+	do_one_branch_coverage(r2n, dt, p, __dump_one_branch_coverage);
+}
+
 struct branch_cnts {
 	unsigned long ok;
 	unsigned long nt;
@@ -825,7 +967,7 @@ void dump_branch_coverage(struct range_to_name *r2n, struct html_out_data *dt)
 	long i, uk_bcnt;
 	struct branch_cnts bcs = {0,0,0,0,0};
 	struct path *p;
-	unsigned long n_all_branch;
+	unsigned long n_all_branch, n_all_uk;
 
 	for (i = 0; i < r2n->cnt; i++) {
 		p = r2n->pt[i];
@@ -845,14 +987,19 @@ void dump_branch_coverage(struct range_to_name *r2n, struct html_out_data *dt)
 		else
 			bcs.ht++;
 	}
-	n_all_branch = (bcs.ok + bcs.ht + bcs.nt) * 2
-		       			+ bcs.uk_through + bcs.uk_not_through;
+	n_all_branch = (bcs.ok + bcs.ht + bcs.nt) * 2;
+	n_all_uk = bcs.uk_through + bcs.uk_not_through;
 	if (HTML_OUTPUT)
 		out_summary_html_branch(dt, bcs.ok, bcs.uk_through, bcs.ht,
-					bcs.nt, n_all_branch);
-	else
-		printf("------ branch coverage (OK:%ld,UK:%ld,HT:%ld,NT:%ld / %ld)\n",
-		       bcs.ok, bcs.uk_through, bcs.ht, bcs.nt, n_all_branch);
+					bcs.nt, n_all_branch, n_all_uk);
+	else {
+		printf("------ branch coverage (OK:%ld,HT:%ld,NT:%ld/%ld=%.2f%%",
+		       bcs.ok, bcs.ht, bcs.nt, n_all_branch,
+		       get_percent(bcs.ok * 2 + bcs.ht, n_all_branch));
+		printf(" UK:%ld/%ld=%.2f%%) ------\n",
+		       bcs.uk_through, n_all_uk,
+		       get_percent(bcs.uk_through, n_all_uk));
+	}
 	if (output_summary)
 		return;
 	if (HTML_OUTPUT)
@@ -896,8 +1043,9 @@ void dump_state_coverage(struct range_to_name *r2n, struct html_out_data *dt)
 	if (HTML_OUTPUT)
 		out_summary_html_state(dt, scs.ok, scs.ok + scs.nt);
 	else
-		printf("------ state coverage (%ld/%ld) ------\n",
-		       scs.ok, scs.ok + scs.nt);
+		printf("------ state coverage (%ld/%ld=%.2f%%) ------\n",
+		       scs.ok, scs.ok + scs.nt,
+		       get_percent(scs.ok, scs.ok + scs.nt));
 	if (output_summary)
 		return;
 	if (HTML_OUTPUT)
@@ -926,6 +1074,25 @@ void dump_state_coverage(struct range_to_name *r2n, struct html_out_data *dt)
  *  display coverage
  *-----------------------------------------------------------------------------
  */
+static int f_chk_ifuncs_are_executed(void *__dt, void *user_data)
+{
+	unsigned long addr = (unsigned long)__dt;
+	struct range_to_name *r2n;
+	unsigned long end;
+	struct func_chk *fc;
+
+	if (is_exclude_faddr(addr) ||
+	    !get_func_info(&r2n_info, addr, &end, &r2n))
+		return 0;
+	fc = search_tree((void*)addr, include_fcs, f_cmp_addr2fc);
+	if (!fc || !fc->cnt) {
+		printf("WARN: ");
+		printf_funcname(&r2n->bi, addr - r2n->offset);
+		printf(" was not executed.\n");
+	}
+	return 0;
+}
+
 int dump_coverage(void)
 {
 	int i;
@@ -942,6 +1109,7 @@ int dump_coverage(void)
 		memset(&dt, 0, sizeof(struct html_out_data));
 		dt.srcdir = srcdir;
 		dt.outdir = outdir;
+		dt.limit_by_funcs = LIMIT_BY_FUNCS;
 
 		if (init_html_output(&dt) < 0)
 			return -1;
@@ -950,67 +1118,52 @@ int dump_coverage(void)
 		return -1;
 
 	u_dt.html_out_dt = &dt;
-	if (LIMIT_BY_FUNCS && !HTML_OUTPUT && !output_summary) {
-		print_func_tree(fc_list);
-		for (i = 0; i < r2n_info.r2n_max; i++) {
-			r2n = r2n_info.all_r2n[i];
-			if (r2n) {
-				pack.called = 0;
-				pack.all = 0;
-				pack.r2n = r2n;
-				for_each_node(fc_list, f_cnt_each_fc, &pack);
-				if (!pack.all)
-					continue;
-				dt.name = r2n->basename;
-				printf("====== %s coverage ======\n",
-				       r2n->basename);
-				dump_branch_coverage(r2n, &dt);
-				dump_state_coverage(r2n, &dt);
-			}
+	if (LIMIT_BY_FUNCS) {
+		/* Check if the inculde functions are executed. */
+		for_each_node(include_funcs, f_chk_ifuncs_are_executed, NULL);
+		if (!output_summary)
+			print_func_tree(fc_list, &dt);
+	}
+	for (i = 0; i < r2n_info.r2n_max; i++) {
+		r2n = r2n_info.all_r2n[i];
+		if (!r2n)
+			continue;
+		pack.called = 0;
+		pack.all = 0;
+		pack.r2n = r2n;
+		for_each_node(fc_list, f_cnt_each_fc, &pack);
+		if (!pack.all)
+			continue;
+		dt.name = r2n->basename;
+		if (HTML_OUTPUT) {
+			out_summary_html_name(&dt);
+			out_summary_html_func(&dt, pack.called, pack.all);
+			out_func_html_start(&dt);
+		} else {
+			printf("====== %s coverage ======\n", r2n->basename);
+			printf("------ function coverage (%ld/%ld=%.2f%%) ------\n",
+			       pack.called, pack.all,
+			       get_percent(pack.called, pack.all));
 		}
-	} else {
-		for (i = 0; i < r2n_info.r2n_max; i++) {
-			r2n = r2n_info.all_r2n[i];
-			if (r2n) {
-				/* dump_func_coverage(r2n) */
-				pack.called = 0;
-				pack.all = 0;
-				pack.r2n = r2n;
-				for_each_node(fc_list, f_cnt_each_fc, &pack);
-				if (!pack.all)
-					continue;
-				dt.name = r2n->basename;
-				if (HTML_OUTPUT) {
-					out_summary_html_name(&dt);
-					out_summary_html_func(&dt,
-							      pack.called, pack.all);
-					out_func_html_start(&dt);
-				} else {
-					printf("====== %s coverage ======\n",
-					       r2n->basename);
-					printf("------ function coverage (%ld/%ld) ------\n",
-					       pack.called, pack.all);
-				}
-				if (!output_summary) {
-					u_dt.r2n = r2n;
-					for_each_node(fc_list, f_dump_each_fc, &u_dt);
-				}
-				if (HTML_OUTPUT)
-					out_func_html_end(&dt);
-
-				dump_branch_coverage(r2n, &dt);
-				dump_state_coverage(r2n, &dt);
-			}
+		if (!output_summary) {
+			u_dt.r2n = r2n;
+			if (LIMIT_BY_FUNCS)
+				get_excluded_tcnt_func_list(&fc_list);
+			for_each_node(fc_list, f_dump_each_fc, &u_dt);
 		}
+		if (HTML_OUTPUT)
+			out_func_html_end(&dt);
+		dump_branch_coverage(r2n, &dt);
+		dump_state_coverage(r2n, &dt);
 	}
 	if (HTML_OUTPUT) {
-		if (exit_html_output(&dt) < 0)
+		if (exit_html_output(&dt, LIMIT_BY_FUNCS) < 0)
 			return -1;
 	}
 	return 0;
 }
 
-int proc_logfile(char *logfile, int id)
+int proc_logfile(char *logfile, int id, char *uname_r)
 {
 	int fd, rc = -1, mode;
 	off_t size;
@@ -1030,7 +1183,7 @@ int proc_logfile(char *logfile, int id)
 				       PM_WITH_PT, mode) < 0)
 				goto EXIT;
 		}
-		if (parse_modules(&r2n_info, dir, PM_WITH_PT, mode) < 0)
+		if (parse_modules(&r2n_info, dir, uname_r, PM_WITH_PT, mode) <0)
 			goto EXIT;
  		r2n_info.from_is_next = chk_from_is_next(dir);
  
@@ -1070,13 +1223,14 @@ void usage(void)
 	fprintf(stderr, "      This option excludes only included function by -I option.\n");
 	fprintf(stderr, "  -S: source directory\n");
 	fprintf(stderr, "  -o: html output directory\n");
-	fprintf(stderr, "      -o and -S options could not use with -s option.\n");
+	fprintf(stderr, "      -o and -S options cannot be used with -s option.\n");
 }
 
 int main(int argc, char *argv[])
 {
 	int i, opt_index, n_logfs = 0;
 	char c, *logfs_top, *prev, *cur;
+	char *uname_r = NULL;
 	unsigned long begin, end;
 	struct option long_options[] = {
 		{"usr", no_argument, NULL, 0},
@@ -1086,7 +1240,7 @@ int main(int argc, char *argv[])
 
 	if (alloc_pid_range(ALL_PID) < 0)
 		err_exit();
-	while ((c = getopt_long(argc, argv, "sa:d:f:I:E:S:o:v", long_options,
+	while ((c = getopt_long(argc, argv, "sa:d:f:I:E:S:o:u:v", long_options,
 				&opt_index)) != -1) {
 		switch (c) {
 		case 0:
@@ -1142,6 +1296,9 @@ int main(int argc, char *argv[])
 			else
 				outdir = optarg;
 			break;
+		case 'u'://////// TEMPORARY CHANGE
+			uname_r = optarg;
+			break;
 		case 'v':
 			verbose = 1;
 			break;
@@ -1150,7 +1307,8 @@ int main(int argc, char *argv[])
 			err_exit();
 		}
 	}
-	if (optind < argc || !n_logfs || (output_summary && outdir)) {
+	if (optind < argc || !n_logfs || (output_summary && outdir) ||
+	    (!includes && excludes)) {
 		usage();
 		err_exit();
 	}
@@ -1161,7 +1319,7 @@ int main(int argc, char *argv[])
 	r2n_info.all_r2n = xcalloc(r2n_info.r2n_max,
 				   sizeof(struct range_to_name*));
 	for (cur = logfs_top, i = 0; i < n_logfs; i++) {
-		if (proc_logfile(cur, i) < 0)
+		if (proc_logfile(cur, i, uname_r) < 0)
 			err_exit();
 		cur += strlen(cur) + 1;
 	}
