@@ -5,7 +5,7 @@
 
 /*****************************************************************************/
 /*  bt_hconv.c - coverage output to html converter                           */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2006                         */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2007                         */
 /*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
@@ -60,35 +60,50 @@ static long src_info_num;
  *  output html
  *-----------------------------------------------------------------------------
  */
+#define TREE_HTML_PATH	"each/__btrax_ftree__.html"
+
 #define COLOR_OK	"#00ff00"
 #define COLOR_HT	"#ffff00"
 #define COLOR_NT	"#ff0000"
 
-static void out_html_header(FILE *f, char *doctype, char *title)
+static void out_html_header(FILE *f, char *doctype, char *title, int use_tree)
 {
 	fprintf(f,
 		"%s\n" \
 		"<html lang=\"en\">\n" \
 		"<head>\n" \
 		"<meta http-equiv=\"Content-Type\" content=\"text/html; charset=ISO-8859-1\">\n" \
-		"<meta http-equiv=\"Content-Style-Type\" content=\"text/css\">\n" \
+		"<meta http-equiv=\"Content-Style-Type\" content=\"text/css\">\n",
+		doctype);
+	if (use_tree) {
+		fprintf(f,
+			"<meta http-equiv=\"Content-Script-Type\" content=\"text/css\">\n" \
+			"<script type=\"text/javascript\"><!--\n" \
+			"function treeMenu(tName) {\n" \
+			"\ttMenu = document.all[tName].style;\n" \
+			"\tif (tMenu.display == 'none') tMenu.display = 'block';\n" \
+			"\telse tMenu.display = 'none'\n" \
+			"}\n" \
+			"--></script>\n");
+	}
+	fprintf(f,
 		"<title>%s</title>\n" \
 		"</head>\n",
-		doctype, title);
+		title);
 }
 
 static void out_html_frame_header(FILE *f, char *title)
 {
 	out_html_header(f,
 			"<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Frameset//EN\">",
-			title);
+			title, 0);
 }
 
-static void out_html_normal_header(FILE *f, char *title)
+static void out_html_normal_header(FILE *f, char *title, int use_tree)
 {
 	out_html_header(f,
 			"<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">",
-			title);
+			title, use_tree);
 }
 
 static int out_top_html(struct html_out_data *dt)
@@ -133,7 +148,7 @@ static int out_title_html(struct html_out_data *dt)
 		fprintf(stderr, "%s can't open.(%s)\n", buf, strerror(errno));
 		return -1;
 	}
-	out_html_normal_header(f, "title");
+	out_html_normal_header(f, "title", 0);
 	fprintf(f,
 		"<body style=\"background-color:silver\">\n" \
 		"<p style=\"font-size:16pt; font-weight:bold; text-align:center\">\n" \
@@ -174,7 +189,7 @@ static void close_html(struct html_out_data *dt)
 
 static void out_summary_html_start(struct html_out_data *dt)
 {
-	out_html_normal_header(dt->summary_out, "summary");
+	out_html_normal_header(dt->summary_out, "summary", 0);
 	fprintf(dt->summary_out,
 		"<body>\n" \
 		"<table summary=\"summary\" border=\"1\">\n" \
@@ -195,22 +210,23 @@ void out_summary_html_func(struct html_out_data *dt,
 	rel_path = get_rel_path(dt, 'f');
 	fprintf(dt->summary_out,
 		"    <td align=\"right\"><a href=\"%s\" target=\"each\" title=\"%s\">%.2f%%</a></td>\n",
-		rel_path, rel_path, (double)n_func * 100 / n_func_all);
+		rel_path, rel_path, get_percent(n_func, n_func_all));
 }
 
 void out_summary_html_branch(struct html_out_data *dt,
 			     long n_br_ok, long n_br_uk,
-			     long n_br_ht, long n_br_nt, long n_br_all)
+			     long n_br_ht, long n_br_nt,
+			     long n_br_all, long n_uk_all)
 {
 	char *rel_path;
 
 	rel_path = get_rel_path(dt, 'b');
 	if (n_br_all)
 		fprintf(dt->summary_out,
-			"    <td align=\"right\"><a href=\"%s\" target=\"each\" title=\"%s\">%.2f%%</a></td>\n",
+			"    <td align=\"right\"><a href=\"%s\" target=\"each\" title=\"%s\">%.2f%%(%.2f%%)</a></td>\n",
 			rel_path, rel_path,
-			(double)(n_br_ok * 2 + n_br_uk + n_br_ht) * 100
-								/ n_br_all);
+			get_percent(n_br_ok * 2 + n_br_ht, n_br_all),
+			get_percent(n_br_uk, n_uk_all));
 	else
 		fprintf(dt->summary_out,
 			"    <td align=\"right\">----</td>\n");
@@ -223,78 +239,93 @@ void out_summary_html_state(struct html_out_data *dt, long n_ok, long n_states)
 	rel_path = get_rel_path(dt, 's');
 	fprintf(dt->summary_out,
 		"    <td align=\"right\"><a href=\"%s\" target=\"each\" title=\"%s\">%.2f%%</a></td></tr>\n",
-		rel_path, rel_path, (double)n_ok * 100 / n_states);
+		rel_path, rel_path, get_percent(n_ok, n_states));
 }
 
-static void out_summary_html_end(struct html_out_data *dt)
+static void out_summary_html_end(struct html_out_data *dt, int limit_by_funcs)
 {
-	fprintf(dt->summary_out,
-		"</table>\n" \
-		"</body></html>\n");
+	fprintf(dt->summary_out, "</table>\n");
+	if (limit_by_funcs)
+		fprintf(dt->summary_out,
+			"<a href=\"%s\" target=\"each\" title=\"ftree\">View function tree.</a>",
+			TREE_HTML_PATH);
+	fprintf(dt->summary_out, "</body></html>\n");
 }
 
 static struct src_info* chk_src_line_type(struct html_out_data*, const char*,
 					  long, char);
 static char* get_src_html_path(struct html_out_data*, struct src_info*, int);
 
-void out_func_html_start(struct html_out_data *dt)
+static int get_fname_and_ref_str(struct html_out_data *dt,
+				 struct range_to_name *r2n,
+				 unsigned long addr, int type,
+				 char buf[MAX_LINE_LEN],
+				 char **p_fname, char **p_ref)
 {
-	char title[MAX_LINE_LEN + 1];
+	const char *src, *func, *tmp;
+	int ln, rc, len, left;
+	size_t offset;
+	struct src_info *info;
+	char *ref = NULL, *p;
 
-	open_html(dt);
-	title[MAX_LINE_LEN] = '\0';
-	sprintf(title, "%s function coverage", dt->name);
-	out_html_normal_header(dt->cur_out, title);
-	fprintf(dt->cur_out,
-		"<body>\n" \
-		"<table summary=\"%s\" border=\"1\"><tr><th>function</th><th>count</th></tr>\n",
-		title);
+	rc = addr_to_func_name_and_offset(&r2n->bi, addr, &func, &offset);
+	if (rc >= 0) {
+		rc = get_source_info(&r2n->bi, addr, &src, &tmp, &ln);
+		if (src && ln && (info = chk_src_line_type(dt, src, ln, type)))
+			ref = get_src_html_path(dt, info, 1);
+		if (offset)
+			len = snprintf(buf, MAX_LINE_LEN, "%s+0x%x",
+				       func, offset);
+		else
+			len = snprintf(buf, MAX_LINE_LEN, "%s", func);
+	} else
+		len = snprintf(buf, MAX_LINE_LEN, "0x%08lx", addr);
+	if (len >= MAX_LINE_LEN)
+		goto ERR;
+	*p_fname = buf;
+	*p_ref = NULL;
+	if (!ref)
+		return 0;
+	p = buf + len + 1;
+	left = MAX_LINE_LEN - len - 1;
+	len = snprintf(p, left, "%s#%d", ref, ln);
+	if (len >= left)
+		goto ERR;
+	*p_ref = p;
+	return 0;
+ERR:
+	fprintf(stderr, "MAX_LINE_LEN too short.\n");
+	return -1;
 }
 
 void out_func_html_each(struct html_out_data *dt, struct range_to_name *r2n,
-			unsigned long addr, long cnt)
+			struct func_chk *fc)
 {
-	const char *src, *func, *tmp;
-	int ln, rc;
-	size_t offset;
-	struct src_info *info;
-	char *color, *ref;
-	char type;
+	char *color, buf[MAX_LINE_LEN], *p_fname, *p_ref;
+	int type;
+	unsigned long addr = fc->addr - r2n->offset;
 
-	if (cnt) {
+	if (fc->cnt) {
 		type = T_EXECUTED;
 		color = COLOR_OK;
 	} else {
 		type = T_NOT_EXECUTED;
 		color = COLOR_NT;
 	}
+	if (get_fname_and_ref_str(dt, r2n, addr, type, buf, &p_fname, &p_ref)
+	    < 0)
+		return;
 	fprintf(dt->cur_out, "<tr><td bgcolor=\"%s\">", color);
-	rc = addr_to_func_name_and_offset(&r2n->bi, addr, &func, &offset);
-	if (rc >= 0) {
-		rc = get_source_info(&r2n->bi, addr, &src, &tmp, &ln);
-		if (src && (info = chk_src_line_type(dt, src, ln, type))) {
-			ref = get_src_html_path(dt, info, 1);
-			if (offset)
-				fprintf(dt->cur_out,
-					"<a href=\"%s#%d\" target=\"src\">%s+0x%x</a></td><td align=\"right\">",
-					ref, ln, func, offset);
-			else
-				fprintf(dt->cur_out,
-					"<a href=\"%s#%d\" target=\"src\">%s</a></td><td align=\"right\">",
-					ref, ln, func);
-		} else {
-			if (offset)
-				fprintf(dt->cur_out,
-					"%s+0x%x</td><td align=\"right\">",
-					func, offset);
-			else
-				fprintf(dt->cur_out,
-					"%s</td><td align=\"right\">", func);
-		}
-	} else
-		fprintf(dt->cur_out,
-			"0x%08lx</td><td align=\"right\">", addr);
-	fprintf(dt->cur_out, "%ld</td></tr>\n", cnt);
+	if (p_ref)
+		fprintf(dt->cur_out, "<a href=\"%s\" target=\"src\">%s</a>",
+			p_ref, p_fname);
+	else
+		fprintf(dt->cur_out, "%s", p_fname);
+	fprintf(dt->cur_out, "</td><td align=\"right\">%ld</td>", fc->cnt);
+	if (dt->limit_by_funcs)
+		fprintf(dt->cur_out, "<td align=\"right\">%ld</td>",
+			fc->excluded_tree_cnt);
+	fprintf(dt->cur_out, "</tr>\n");
 }
 
 void out_func_html_end(struct html_out_data *dt)
@@ -312,7 +343,7 @@ void out_branch_html_start(struct html_out_data *dt)
 	open_html(dt);
 	title[MAX_LINE_LEN] = '\0';
 	sprintf(title, "%s branch coverage", dt->name);
-	out_html_normal_header(dt->cur_out, title);
+	out_html_normal_header(dt->cur_out, title, 0);
 	fprintf(dt->cur_out,
 		"<body>\n" \
 		"<table summary=\"%s\" border=\"1\"><tr>\n" \
@@ -336,7 +367,7 @@ static struct src_info* chk_branch_point(struct html_out_data *dt,
 
 	*__src = NULL;
 	rc = get_source_info(&r2n->bi, addr, &src, &func, ln);
-	if (rc >= 0 && src) {
+	if (rc >= 0 && src && ln) {
 		*__src = strdup(src);
 		type = cnt ? T_EXECUTED : T_NOT_EXECUTED;
 		return chk_src_line_type(dt, src, *ln, type);
@@ -344,9 +375,12 @@ static struct src_info* chk_branch_point(struct html_out_data *dt,
 	return NULL;
 }
 
-void out_branch_html_each(struct html_out_data *dt, struct range_to_name *r2n,
-			  unsigned long base, unsigned long branch, long b_cnt,
-			  unsigned long fall, long f_cnt)
+static inline void __out_branch_html_each(FILE *out, struct html_out_data *dt,
+					  struct range_to_name *r2n,
+					  unsigned long base,
+					  unsigned long branch,
+					  long b_cnt, unsigned long fall,
+					  long f_cnt)
 {
 	char *src, *color, *ref = "";
 	int ln;
@@ -358,56 +392,58 @@ void out_branch_html_each(struct html_out_data *dt, struct range_to_name *r2n,
 	else if (b_cnt || f_cnt)
 		color = COLOR_HT;
 
+	fprintf(out, "<td bgcolor=\"%s\">", color);
 	info = chk_branch_point(dt, r2n, base, b_cnt + f_cnt, &src, &ln);
 	if (info) {
 		ref = get_src_html_path(dt, info, 1);
-		fprintf(dt->cur_out,
-			"<tr><td bgcolor=\"%s\"><a href=\"%s#%d\" target=\"src\">%s,%d</a></td>\n",
-			color, ref, ln, basename(src), ln);
+		fprintf(out,
+			"<a href=\"%s#%d\" target=\"src\">%s,%d</a></td>\n",
+			ref, ln, basename(src), ln);
 		free(src);
 	} else if (src && ln) {
-		fprintf(dt->cur_out,
-			"<tr><td bgcolor=\"%s\">%s,%d</td>\n",
-			color, basename(src), ln);
+		fprintf(out, "%s,%d</td>\n", basename(src), ln);
 	} else {
-		fprintf(dt->cur_out,
-			"<tr><td bgcolor=\"%s\">0x%08lx</td>\n", color, base);
+		fprintf(out, "0x%08lx</td>\n", base);
 	}
 	if (branch == UNKNOWN_BADDR)
-		fprintf(dt->cur_out,
-			"    <td>----------</td><td align=\"right\">%ld</td>\n",
-			b_cnt);
+		fprintf(out, "    <td>----------</td>");
 	else {
 		chk_branch_point(dt, r2n, branch, b_cnt, &src, &ln);
 		if (src && ln) {
-			fprintf(dt->cur_out,
-				"    <td>%s,%d</td><td align=\"right\">%ld</td>\n",
-				basename(src), ln, b_cnt);
+			fprintf(out, "    <td>%s,%d</td>",
+				basename(src), ln);
 			free(src);
 		} else {
-			fprintf(dt->cur_out,
-				"    <td>0x%08lx</td><td align=\"right\">%ld</td>\n",
-				branch, b_cnt);
+			fprintf(out, "    <td>0x%08lx</td>", branch);
 		}
 	}
+	fprintf(out, "<td align=\"right\">%ld</td>\n", b_cnt);
 	if (fall == UNKNOWN_BADDR)
-		fprintf(dt->cur_out,
-			"    <td></td><td align=\"right\"></td></tr>\n");
+		fprintf(out, "    <td></td><td align=\"right\"></td>");
 	else {
 		chk_branch_point(dt, r2n, fall, f_cnt, &src, &ln);
 		if (src && ln) {
-			fprintf(dt->cur_out,
-				"    <td>%s,%d</td><td align=\"right\">%ld</td></tr>\n",
-				basename(src), ln, f_cnt);
+			fprintf(out, "    <td>%s,%d</td>",
+				basename(src), ln);
 			free(src);
 		} else {
-			fprintf(dt->cur_out,
-				"    <td>0x%08lx</td><td align=\"right\">%ld</td></tr>\n",
-				fall, f_cnt);
+			fprintf(out, "    <td>0x%08lx</td>", fall);
 		}
+		fprintf(out, "<td align=\"right\">%ld</td>",
+			f_cnt);
 	}
 }
 
+void out_branch_html_each(struct html_out_data *dt, struct range_to_name *r2n,
+			  unsigned long base, unsigned long branch,
+			  long b_cnt, unsigned long fall, long f_cnt)
+{
+	fprintf(dt->cur_out, "<tr>");
+	__out_branch_html_each(dt->cur_out, dt, r2n,
+			       base, branch, b_cnt, fall, f_cnt);
+	fprintf(dt->cur_out, "</tr>\n");
+}
+
 void out_branch_html_end(struct html_out_data *dt)
 {
 	fprintf(dt->cur_out,
@@ -423,15 +459,16 @@ void out_state_html_start(struct html_out_data *dt)
 	open_html(dt);
 	title[MAX_LINE_LEN] = '\0';
 	sprintf(title, "%s state coverage", dt->name);
-	out_html_normal_header(dt->cur_out, title);
+	out_html_normal_header(dt->cur_out, title, 0);
 	fprintf(dt->cur_out,
 		"<body>\n" \
 		"<table summary=\"%s\" border=\"1\"><tr><th>state</th><th>exec</th></tr>\n",
 		title);
 }
 
-void out_state_html_each(struct html_out_data *dt, struct range_to_name *r2n,
-			 int is_exec, unsigned long addr)
+static inline void __out_state_html_each(FILE *out, struct html_out_data *dt,
+					 struct range_to_name *r2n,
+					 int is_exec, unsigned long addr)
 {
 	int ln;
 	const char *src, *func;
@@ -447,25 +484,220 @@ void out_state_html_each(struct html_out_data *dt, struct range_to_name *r2n,
 		color = COLOR_NT;
 		mark = "N";
 	}
+	fprintf(out, "<td bgcolor=\"%s\">", color);
 	get_source_info(&r2n->bi, addr, &src, &func, &ln);
 	if (ln && (info = chk_src_line_type(dt, src, ln, type))) {
 		ref = get_src_html_path(dt, info, 1);
 		p = strdup(src);
-		fprintf(dt->cur_out,
-			"<tr><td bgcolor=\"%s\"><a href=\"%s#%d\" target=\"src\">%s,%d</a></td><td align=\"center\">%s</td></tr>\n",
-			color, ref, ln, basename(p), ln, mark);
+		fprintf(out, "<a href=\"%s#%d\" target=\"src\">%s,%d</a>",
+			ref, ln, basename(p), ln);
 		free(p);
 	} else if (src && ln) {
 		p = strdup(src);
-		fprintf(dt->cur_out,
-			"<tr><td bgcolor=\"%s\">%s,%d</td><td align=\"center\">%s</td></tr>\n",
-			color, basename(p), ln, mark);
+		fprintf(out, "%s,%d", basename(p), ln);
 		free(p);
 	} else {
-		fprintf(dt->cur_out,
-			"<tr><td bgcolor=\"%s\">0x%08lx</td><td align=\"center\">%s</td></tr>\n",
-			color, addr, mark);
+		fprintf(out, "0x%08lx", addr);
 	}
+	fprintf(out, "</td><td align=\"center\">%s</td>", mark);
+}
+
+void out_state_html_each(struct html_out_data *dt, struct range_to_name *r2n,
+			 int is_exec, unsigned long addr)
+{
+	fprintf(dt->cur_out, "<tr>");
+	__out_state_html_each(dt->cur_out, dt, r2n, is_exec, addr);
+	fprintf(dt->cur_out, "</tr>\n");
+}
+
+#define TREE_MARGIN_VAL		2.0
+#define TREE_CTRL_CHILD_CHAR	'T'
+#define TREE_CTRL_COV_CHAR	'C'
+
+void out_func_tree_html_start(struct html_out_data *dt,
+			      char *s_inc, char *s_exc)
+{
+	char path[PATH_MAX + 1];
+	char title[MAX_LINE_LEN + 1];
+
+	path[PATH_MAX] = '\0';
+	snprintf(path, PATH_MAX, "%s/%s", dt->outdir, TREE_HTML_PATH);
+	dt->ftree_out = fopen(path, "w");
+	if (!dt->ftree_out) {
+		fprintf(stderr, "%s can't open.(%s)\n", path,
+			strerror(errno));
+		return;
+	}
+	title[MAX_LINE_LEN] = '\0';
+	sprintf(title, "function tree");
+	out_html_normal_header(dt->ftree_out, title, 1);
+	fprintf(dt->ftree_out, "<body>\n");
+	fprintf(dt->ftree_out, "<b>includes: </b>%s<br>\n", s_inc);
+	fprintf(dt->ftree_out, "<b>excludes: </b>%s<br>\n", s_exc);
+	fprintf(dt->ftree_out,
+		"<a href=\"javascript:treeMenu('cov_format')\">" \
+		"View coverage table format</a>\n" \
+		"<div id=\"cov_format\" style=\"display:none\">" \
+		"<table summary=\"cov_format\" border=\"1\">" \
+		"<tr>" \
+		"<th colspan=\"2\">state coverage</th>" \
+		"<th colspan=\"5\">branch coverage</th>" \
+		"</tr>" \
+		"<tr>" \
+		"<th>state</th><th>exec</th>" \
+		"<th>base</th><th>branch</th><th>cnt</th>" \
+		"<th>next</th><th>cnt</th>" \
+		"</tr>" \
+		"</table></div><br><br>\n");
+}
+
+static inline void out_ftree_html_each_bcov(struct range_to_name *r2n,
+					    struct html_out_data *dt,
+					    int bcov_type, unsigned long base,
+					    unsigned long branch, long b_cnt,
+					    unsigned long fall, long f_cnt)
+{
+	__out_branch_html_each(dt->ftree_out, dt, r2n,
+			       base, branch, b_cnt, fall, f_cnt);
+}
+
+static void out_ftree_html_each_cov(struct html_out_data *dt,
+				    struct range_to_name *r2n,
+				    char *fname, struct func_chk *fc, int nest,
+				    int type, int has_child)
+{
+	struct path *p;
+	long i;
+
+	fprintf(dt->ftree_out,
+		"<div id=\"%s_%c\" style=\"display:none\">\n" \
+		"<table summary=\"%s\" border=\"1\" style=\"margin-left: %.1fem\">\n",
+		fname, TREE_CTRL_COV_CHAR, fname, TREE_MARGIN_VAL);
+	for (i = fc->i; i < r2n->cnt; i++) {
+		p = r2n->pt[i];
+		if (p->next > fc->end)
+			break;
+		fprintf(dt->ftree_out, "<tr>");
+		__out_state_html_each(dt->ftree_out, dt, r2n, p->cnt > 0,
+				      p->addr);
+		if (p->type == BTYPE_BRANCH || IS_SWITCH_JMP(p))
+			do_one_branch_coverage(r2n, dt, p,
+					       out_ftree_html_each_bcov);
+		else
+			fprintf(dt->ftree_out,
+				"<td></td><td></td><td></td><td></td><td></td>");
+		fprintf(dt->ftree_out, "</tr>\n");
+	}
+	fprintf(dt->ftree_out, "</table><br></div>");
+}
+
+void out_func_tree_html_each_enter(struct html_out_data *dt,
+				   struct range_to_name *r2n,
+				   struct func_chk *fc, int nest, int type,
+				   int has_child)
+{
+	unsigned long addr = fc->addr - r2n->offset;
+	char *color, buf[MAX_LINE_LEN], *p_fname, *p_ref;
+	int exec_type;
+
+	switch (type) {
+	case BCOV_TYPE_OK:
+		exec_type = T_EXECUTED;
+		color = COLOR_OK;
+		break;
+	case BCOV_TYPE_NT:
+		exec_type = T_NOT_EXECUTED;
+		color = COLOR_NT;
+		break;
+	case BCOV_TYPE_HT:
+		exec_type = T_NOT_EXECUTED;
+		color = COLOR_HT;
+		break;
+	}
+	if (get_fname_and_ref_str(dt, r2n, addr, exec_type, buf,
+				  &p_fname, &p_ref) < 0)
+		return;
+
+	fprintf(dt->ftree_out,
+		"<div style=\"margin-left: %.1fem\"><nobr>",
+		nest ? TREE_MARGIN_VAL : 0);
+	if (!has_child)
+		fprintf(dt->ftree_out, "%c ", TREE_CTRL_CHILD_CHAR);
+	else
+		fprintf(dt->ftree_out,
+			"<a href=\"javascript:treeMenu('%s_%c')\">%c</a> ",
+			p_fname, TREE_CTRL_CHILD_CHAR, TREE_CTRL_CHILD_CHAR);
+	if (type == BCOV_TYPE_HT)
+		fprintf(dt->ftree_out, "%c ", TREE_CTRL_COV_CHAR);
+	else
+		fprintf(dt->ftree_out,
+			"<a href=\"javascript:treeMenu('%s_%c')\">%c</a> ",
+			p_fname, TREE_CTRL_COV_CHAR, TREE_CTRL_COV_CHAR);
+	fprintf(dt->ftree_out, "<span style=\"background:%s\">", color);
+	if (p_ref)
+		fprintf(dt->ftree_out, "<a href=\"%s\" target=\"src\">%s</a>",
+			p_ref, p_fname);
+	else
+		fprintf(dt->ftree_out, "%s", p_fname);
+	fprintf(dt->ftree_out, "</span> %ld (F:%ld)</nobr>",
+		fc->cnt, fc->excluded_tree_cnt);
+
+	if (type != BCOV_TYPE_HT)
+		out_ftree_html_each_cov(dt, r2n, p_fname, fc, nest, type,
+					has_child);
+	if (has_child)
+		fprintf(dt->ftree_out,
+			"<div id=\"%s_%c\" style=\"display:block\">\n",
+			p_fname, TREE_CTRL_CHILD_CHAR);
+}
+
+void out_func_tree_html_each_exit(struct html_out_data *dt, int nest,
+				  int has_child)
+{
+	if (has_child)
+		fprintf(dt->ftree_out, "</div>");
+	fprintf(dt->ftree_out, "</div>\n");
+}
+
+void out_func_tree_html_each_invalid(struct html_out_data *dt,
+				     struct range_to_name *r2n,
+				     struct func_chk *fc)
+{
+	unsigned long addr = fc->addr - r2n->offset;
+	char buf[MAX_LINE_LEN], *p_fname, *p_ref;
+
+	if (get_fname_and_ref_str(dt, r2n, addr, T_EXECUTED, buf,
+				  &p_fname, &p_ref) < 0)
+		return;
+	fprintf(dt->ftree_out, "UT <span style=\"background:%s\">", COLOR_OK);
+	if (p_ref)
+		fprintf(dt->ftree_out, "<a href=\"%s\" target=\"src\">%s</a>",
+			p_ref, p_fname);
+	else
+		fprintf(dt->ftree_out, "%s", p_fname);
+	fprintf(dt->ftree_out, "</span> %ld</nobr><br>",
+		fc->cnt);
+}
+
+void out_func_tree_html_end(struct html_out_data *dt)
+{
+	fprintf(dt->ftree_out,
+		"</body></html>\n");
+	fclose(dt->ftree_out);
+}
+
+void out_func_html_start(struct html_out_data *dt)
+{
+	char title[MAX_LINE_LEN + 1];
+
+	open_html(dt);
+	title[MAX_LINE_LEN] = '\0';
+	sprintf(title, "%s function coverage", dt->name);
+	out_html_normal_header(dt->cur_out, title, 0);
+	fprintf(dt->cur_out,
+		"<body>\n" \
+		"<table summary=\"%s\" border=\"1\"><tr><th>function</th><th>count</th>%s</tr>\n",
+		title, dt->limit_by_funcs ? "<th>n-function</th>" : "");
 }
 
 static char* get_src_html_path(struct html_out_data *dt, struct src_info *info,
@@ -482,7 +714,7 @@ static char* get_src_html_path(struct html_out_data *dt, struct src_info *info,
 		rc = snprintf(info->html_out_path, PATH_MAX,
 			      "%s/src/%s.html", dt->outdir, info->path + 1);
 	if (rc > PATH_MAX)
-		fprintf(stderr, "WARN: PATH_MAX is too short.\n");
+		printf("WARN: PATH_MAX is too short.\n");
 	info->is_ref = is_ref;
 	return info->html_out_path;
 }
@@ -635,19 +867,24 @@ static void out_src_html(struct html_out_data *dt)
 			break;
 		}
 		ln_cols = get_ln_cols(info);
-		out_html_normal_header(w, "src");
+		out_html_normal_header(w, "src", 0);
 		fprintf(w, "<body style=\"font-size:8pt\"><pre>\n");
 		type = T_NONE;
 		for (ln = 0; fgets(buf, MAX_LINE_LEN, r); ln++) {
-			//printf("%*ld: %d\n", ln_cols, ln + 1, info->exec_types[ln]);
+			/* for DEBUG
+			printf("%*ld: %d\n",
+			       ln_cols, ln + 1, info->exec_types[ln]);
+			       */
 			fprintf(w, "%*ld: ", ln_cols, ln + 1);
 			if (info->exec_types[ln] != T_NONE)
 				type = info->exec_types[ln];
 			buf[strlen(buf) - 1] = '\0';
 			escape_for_html(buf);
-			fprintf(w, "<a name=\"%ld\" style=\"background:%s\">%s</a>\n",
-				ln + 1, type2color(type), buf);
-
+			if (type == T_NONE)
+				fprintf(w, "%s\n", buf);
+			else
+				fprintf(w, "<a name=\"%ld\" style=\"background:%s\">%s</a>\n",
+					ln + 1, type2color(type), buf);
 			/* '}' of a line-head is regarded as the end of a
 			 * function.
 			 */
@@ -719,10 +956,10 @@ int init_html_output(struct html_out_data *dt)
 	return 0;
 }
 
-int exit_html_output(struct html_out_data *dt)
+int exit_html_output(struct html_out_data *dt, int limit_by_funcs)
 {
 	out_src_html(dt);
-	out_summary_html_end(dt);
+	out_summary_html_end(dt, limit_by_funcs);
 	if (dt->summary_out)
 		fclose(dt->summary_out);
 	return 0;
