@@ -5,7 +5,7 @@
 
 /*****************************************************************************/
 /*  bt_execpath.c - execution path display program                           */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2006                         */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2007                         */
 /*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
@@ -30,7 +30,7 @@
 #include "bt_utils.h"
 
 #define BT_EXECPATH_VER	BT_VER
-#define	COPYRIGHT	"Copyright (c) Hitachi, Ltd. 2005-2006"
+#define	COPYRIGHT	"Copyright (c) Hitachi, Ltd. 2005-2007"
 
 struct proc_each_rec_data {
 	struct addr_range	*range;
@@ -65,7 +65,7 @@ int print_exec_line(unsigned long addr, struct range_to_name *r2n)
 	addr -= r2n->offset;
 	printf("%s\t0x%08lx\t", r2n->basename, addr);
 	rc = get_source_info(bi, addr, &src_name, &func_name, &lno);
-	if (rc >= 0 && src_name) {
+	if (rc >= 0 && src_name && lno) {
 		addr_to_func_name_and_offset(bi, addr, &func_name, &offset);
 		printf_func_name(func_name, offset);
 		if (src_name)
@@ -155,10 +155,10 @@ int proc_each_record_for_normal(struct bt_record *__p, off_t i_rec, void *tmp)
 
 	if (is_warn_record(&p)) {
 		struct warn_record *warn = (struct warn_record*)&p;
-		fprintf(stderr, "WARN: bts left only: %ld\n", warn->left);
+		printf("WARN: bts left only: %ld\n", warn->left);
 		return 0;
 	}
-	if (is_pid_record(&p))
+	if (is_pid_record(&p) || is_comm_record(&p))
 		return 0;
 	rc = chk_fix_from_cache(&r2n_info, &p.from, p.to, NULL, NULL, NULL);
 	if (rc < 0)
@@ -205,7 +205,7 @@ void dump_ret_chk(int n)
 void set_expect_ret_addr(int type, unsigned long addr)
 {
 	if (p_ret_chk_last - ret_chk >= MAX_NEST) {
-		fprintf(stderr, "WARN: MAX_NEST is too small.\n");
+		printf("WARN: MAX_NEST is too small.\n");
 		return;
 	}
 	p_ret_chk_last->type = type;
@@ -404,10 +404,10 @@ int proc_each_record_for_summary(struct bt_record *__p, off_t i_rec, void *tmp)
 
 	if (!nest_chk && is_warn_record(&p)) {
 		struct warn_record *warn = (struct warn_record*)&p;
-		fprintf(stderr, "WARN: bts left only: %ld\n", warn->left);
+		printf("WARN: bts left only: %ld\n", warn->left);
 		return 0;
 	}
-	if (is_pid_record(&p))
+	if (is_pid_record(&p) || is_comm_record(&p))
 		return 0;
 	r2n_to = addr_to_r2n(&r2n_info, p.to);
 	rc = chk_fix_from_cache(&r2n_info, &p.from, p.to, &r2n, &p_from, NULL);
@@ -637,7 +637,7 @@ int proc_logfile(char *logfile)
 			       PM_WITH_PT, PM_VERBOSE) <0)
 			goto EXIT;
 	}
-	if (parse_modules(&r2n_info, dir, PM_WITH_PT, PM_VERBOSE) < 0)
+	if (parse_modules(&r2n_info, dir, NULL, PM_WITH_PT, PM_VERBOSE) < 0)
 		goto EXIT;
 	r2n_info.from_is_next = chk_from_is_next(dir);
 
