@@ -5,7 +5,7 @@
 
 /*****************************************************************************/
 /*  bt_ar_parse.c - /proc/modules and /proc/PID/maps parser                  */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2006                         */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2007                         */
 /*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
@@ -110,21 +110,178 @@ static int is_same_name_registered(struct r2n_info *r2i, char *name)
 	return 0;
 }
 
+/* Size of the 'buf' must be PATH_MAX */
+static char* get_pt_fpath(char *buf, char *name, char *uname_r, int is_read)
+{
+	int i;
+	struct stat st;
+	char *dirs[] = { "/tmp", "/var/tmp" };
+
+	for (i = 0; i < sizeof(dirs) / sizeof(dirs[0]); i++) {
+		snprintf(buf, PATH_MAX, "%s/btrax.pt.%s-%s",
+			 dirs[i], name, uname_r);
+		if (is_read) {
+			if (lstat(buf, &st) >= 0 && S_ISREG(st.st_mode))
+				return buf;
+		} else {
+			if (lstat(dirs[i], &st) >= 0 && S_ISDIR(st.st_mode))
+				return buf;
+		}
+	}
+	return NULL;
+}
+
+#if 1
+static int chk_self_and_pt_mtime(char *self, char *pt)
+{
+	struct stat self_st, pt_st;
+
+	if (lstat(self, &self_st) < 0 || lstat(pt, &pt_st) < 0)
+		return -1;
+	if (self_st.st_mtime >= pt_st.st_mtime)
+		return -1;
+	return 0;
+}
+#endif
+
+#define read_dt(fd, dt) \
+	if (read((fd), &(dt), sizeof(dt)) != sizeof(dt)) goto READ_ERR_EXIT
+
+static int read_pt_from_file(struct range_to_name *r2n, char *vmlinux_path,
+			     struct stat *vmlinux_st, char *uname_r)
+{
+	unsigned long i;
+	int fd, len;
+	struct stat st;
+	char ver[16], fpath[PATH_MAX], path[PATH_MAX];
+	struct path *pt;
+
+	if (!get_pt_fpath(fpath, r2n->basename, uname_r, 1))
+		return -1;
+	if ((fd = open(fpath, O_RDONLY)) < 0) {
+		fprintf(stderr, "'%s' open failed.(%s)\n",
+			fpath, strerror(errno));
+		return -1;
+	}
+#if 1
+	/* If this program's modified time is newer than path-tree file,
+	 * there is posibility of updating of binutils.
+	 */
+	if ((len = readlink("/proc/self/exe", path, PATH_MAX)) < 0) {
+		fprintf(stderr, "readlink failed.(%s)\n", strerror(errno));
+		goto ERR_EXIT;
+	}
+	path[len] = '\0';
+	if (chk_self_and_pt_mtime(path, fpath) < 0)
+		return -1;
+#endif
+
+	read_dt(fd, len);
+	if (len >= 16)
+		goto ERR_EXIT;
+	if (read(fd, ver, len) != len)
+		goto READ_ERR_EXIT;
+	ver[len] = '\0';
+	if (strcmp(BT_VER, ver) != 0)
+		goto ERR_EXIT;
+	read_dt(fd, len);
+	if (read(fd, path, len) != len)
+		goto READ_ERR_EXIT;
+	path[len] = '\0';
+	if (strcmp(vmlinux_path, path) != 0)
+		goto ERR_EXIT;
+	read_dt(fd, st.st_size);
+	if (vmlinux_st->st_size != st.st_size)
+		goto ERR_EXIT;
+	read_dt(fd, st.st_mtime);
+	if (vmlinux_st->st_mtime != st.st_mtime)
+		goto ERR_EXIT;
+	read_dt(fd, r2n->cnt);
+	r2n->pt = xmalloc(sizeof(struct path*) * r2n->cnt);
+	for (i = 0; i < r2n->cnt; i++) {
+		pt = xmalloc(sizeof(struct path));
+		read_dt(fd, *pt);
+		r2n->pt[i] = pt;
+	}
+	close(fd);
+	return 0;
+
+READ_ERR_EXIT:
+	fprintf(stderr, "read failed.(%s)\n", strerror(errno));
+ERR_EXIT:
+	close(fd);
+	return -1;
+}
+
+#define write_dt(fd, dt) \
+	if (write((fd), &(dt), sizeof(dt)) < 0) goto ERR_EXIT
+
+static int write_pt_to_file(struct range_to_name *r2n, char *vmlinux_path,
+			    struct stat *st, char *uname_r)
+{
+	char fpath[PATH_MAX];
+	unsigned long i;
+	int fd, len;
+
+	if (!get_pt_fpath(fpath, r2n->basename, uname_r, 0))
+		return -1;
+	if ((fd = creat(fpath, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH))
+	    < 0) {
+		fprintf(stderr, "'%s' open failed.(%s)\n",
+			fpath, strerror(errno));
+		return -1;
+	}
+	/* The 'sticky' bit on a temporary directory is typically set.
+	 * In this case, the 'creat' function can not set the write permission
+	 * for another users. So, we call 'chmod' function.
+	 */
+	chmod(fpath, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH);
+
+	len = strlen(BT_VER);
+	write_dt(fd, len);
+	if (write(fd, BT_VER, len) != len)
+		goto ERR_EXIT;
+	len = strlen(vmlinux_path);
+	write_dt(fd, len);
+	if (write(fd, vmlinux_path, len) != len)
+		goto ERR_EXIT;
+	write_dt(fd, st->st_size);
+	write_dt(fd, st->st_mtime);
+	write_dt(fd, r2n->cnt);
+	for (i = 0; i < r2n->cnt; i++)
+		write_dt(fd, *r2n->pt[i]);
+	close(fd);
+	return 0;
+ERR_EXIT:
+	fprintf(stderr, "write failed.(%s)\n", strerror(errno));
+	ftruncate(fd, 0);
+	close(fd);
+	return -1;
+}
+
 static int prepare_obj_file(struct range_to_name *r2n, char *kallsyms,
-			    int with_pt, int verbose)
+			    char *uname_r, int with_pt, int verbose)
 {
 	char path[MAX_LINE_LEN];
+	int is_kernel = 0;
+	struct stat st;
 
 	snprintf(path, MAX_LINE_LEN, "%s/%s", r2n->dirname, r2n->basename);
 	/* In case of without path-tree, do not print 'checking ...' message */
 	if (verbose)
 		printf("checking %s...\n", path);
+	if (stat(path, &st) < 0) {
+		fprintf(stderr, "'%s' stat failed.(%s)\n",
+			path, strerror(errno));
+		return -1;
+	}
 	if (init_bfd_if(&r2n->bi, strdup(path), kallsyms) < 0)
 		return -1;
 	if (!r2n->begin && !r2n->end) {
 		struct addr_range *range = get_pid_addr_range(ALL_PID);
 		bfd_vma min, max;
 
+		is_kernel = 1;
 		get_kernel_min_max_addr(&r2n->bi, &min, &max);
 		if (is_addr_range_match(min, max, range)) {
 			r2n->begin = min;
@@ -140,16 +297,23 @@ static int prepare_obj_file(struct range_to_name *r2n, char *kallsyms,
 		*/
 	}
 	remove_useless_fsyms(&r2n->bi, r2n->begin, r2n->end, r2n->offset);
-	if (with_pt)
-		return chk_path_tree(&r2n->bi, &r2n->pt, &r2n->cnt,r2n->offset);
-	else
-		return 0;
+	if (with_pt) {
+		if (is_kernel) {
+			if (read_pt_from_file(r2n, path, &st, uname_r) >= 0)
+				return 0;
+		}
+		if (chk_path_tree(&r2n->bi, &r2n->pt, &r2n->cnt,r2n->offset) <0)
+			return -1;
+		if (is_kernel)
+			return write_pt_to_file(r2n, path, &st, uname_r);
+	}
+	return 0;
 }
 
 static int add_range_to_name(struct r2n_info *r2i,
 			     unsigned long begin, unsigned long end,
-			     char *name, char *kallsyms, int is_module,
-			     int with_pt, int verbose)
+			     char *name, char *kallsyms, char *uname_r,
+			     int is_module, int with_pt, int verbose)
 {
 	int rc, i, step = 128, r2n_max;
 	struct range_to_name *r2n, **all_r2n;
@@ -173,7 +337,7 @@ static int add_range_to_name(struct r2n_info *r2i,
 		r2n_max += step;
 	}
 	all_r2n[i] = r2n;
-	rc = prepare_obj_file(r2n, kallsyms, with_pt, verbose);
+	rc = prepare_obj_file(r2n, kallsyms, uname_r, with_pt, verbose);
 	if (rc < 0)
 		//return 0;
 		return -1;
@@ -659,14 +823,10 @@ EXIT:
 	return rc;
 }
 
-static int get_module_full_path(char *buf, int buf_len, const char *name)
+static int get_module_full_path(char *buf, int buf_len, const char *name,
+				char *uname_r)
 {
-	struct utsname utsn;
-
-	if (uname(&utsn) < 0)
-		return -1;
-	if (buf[0] == '\0')
-		snprintf(buf, buf_len, "/lib/modules/%s", utsn.release);
+	snprintf(buf, buf_len, "/lib/modules/%s", uname_r);
 	return find_under_path(buf, buf_len, name, 1);
 }
 
@@ -680,40 +840,34 @@ static int __get_vmlinux_full_path(char *buf, int buf_len)
 	return find_under_path(buf, buf_len, "vmlinux", 1);
 }
 
-static int get_vmlinux_full_path(char *buf, int buf_len)
+static int get_vmlinux_full_path(char *buf, int buf_len, char *uname_r)
 {
-	struct utsname utsn;
-
-	if (buf[0] != '\0') {	/* search start from "buf" path */
-		if (__get_vmlinux_full_path(buf, buf_len) >= 0)
-			return 0;
-	}
-	if (uname(&utsn) < 0)
-		return -1;
-	snprintf(buf, buf_len, "/usr/lib/debug/lib/modules/%s", utsn.release);
+	snprintf(buf, buf_len, "/usr/lib/debug/lib/modules/%s", uname_r);
 	if (__get_vmlinux_full_path(buf, buf_len) >= 0)
 		return 0;
-	snprintf(buf, buf_len, "/lib/modules/%s/build", utsn.release);
+	snprintf(buf, buf_len, "/lib/modules/%s/build", uname_r);
 	if (__get_vmlinux_full_path(buf, buf_len) >= 0)
 		return 0;
 	return -1;
 }
 
-//#define HOME_DEBUG
-#ifdef HOME_DEBUG
-#define DBG_MOD_PATH	"/home/tetsu/siken/to/copied_objcts/modules"
-#endif
-
-int parse_modules(struct r2n_info *r2i, char *dir, int with_pt, int verbose)
+int parse_modules(struct r2n_info *r2i, char *dir, char *uname_r,
+		  int with_pt, int verbose)
 {
 	FILE *fd;
 	int rc, num, len;
 	char path[128], kallsyms[128];
 	char buf[MAX_LINE_LEN + 1], *p, fullname[MAX_LINE_LEN + 1];
+	struct utsname utsn;
 	struct addr_range *range;
 	unsigned long from, to;
 
 	rc = -1;
+	if (!uname_r) {
+		if (uname(&utsn) < 0)
+			goto EXIT;
+		uname_r = utsn.release;
+	}
 	if (!dir)
 		goto CHK_KERNEL;
 	snprintf(kallsyms, sizeof(kallsyms), "%s/kallsyms", dir);
@@ -737,31 +891,24 @@ int parse_modules(struct r2n_info *r2i, char *dir, int with_pt, int verbose)
 		if (!is_addr_range_match(from, to, range))
 			continue;
 		snprintf(buf, MAX_LINE_LEN, "%s.ko", fullname);
-#ifdef HOME_DEBUG
-		sprintf(fullname, "%s", DBG_MOD_PATH);
-#else
 		fullname[0] = '\0';
-#endif
-		if (get_module_full_path(fullname, MAX_LINE_LEN, buf) < 0) {
+		if (get_module_full_path(fullname, MAX_LINE_LEN, buf, uname_r)
+		    < 0) {
 			if (verbose)
 				printf("WARN: %s not found.\n", buf);
 			continue;
 		}
-		if (add_range_to_name(r2i, from, to, fullname, kallsyms, 1,
-				      with_pt, verbose) < 0)
+		if (add_range_to_name(r2i, from, to, fullname, kallsyms,
+				      uname_r, 1, with_pt, verbose) < 0)
 			goto EXIT;
 	}
 CHK_KERNEL:
-#ifdef HOME_DEBUG
-	sprintf(fullname, "%s", DBG_MOD_PATH);
-#else
 	fullname[0] = '\0';
-#endif
-	if (get_vmlinux_full_path(fullname, MAX_LINE_LEN) < 0) {
+	if (get_vmlinux_full_path(fullname, MAX_LINE_LEN, uname_r) < 0) {
 		if (verbose)
 			printf("WARN: vmlinux not found.\n");
 	} else {
-		if (add_range_to_name(r2i, 0, 0, fullname, NULL, 0,
+		if (add_range_to_name(r2i, 0, 0, fullname, NULL, uname_r, 0,
 				      with_pt, verbose) < 0)
 			goto EXIT;
 	}
@@ -846,9 +993,10 @@ int get_func_info(struct r2n_info *r2i, unsigned long begin,
 	for (i = 0; i < r2i->r2n_max; i++) {
 		r2n = r2i->all_r2n[i];
 		if (r2n) {
-			end = r2n->end;
-			if (get_end_of_func(&r2n->bi, begin, &end)) {
-				*p_end = end;
+			end = r2n->end - r2n->offset;
+			if (get_end_of_func(&r2n->bi, begin - r2n->offset,
+					    &end)) {
+				*p_end = end + r2n->offset;
 				*p_r2n = r2n;
 				return 1;
 			}
@@ -902,7 +1050,7 @@ int parse_maps(struct r2n_info *r2i, char *dir, char *fname,
 		to -= 1;
 		if (!is_addr_range_match(from, to, range))
 			continue;
-		if (add_range_to_name(r2i, from, to, p, NULL, 0,
+		if (add_range_to_name(r2i, from, to, p, NULL, NULL, 0,
 				      with_pt, verbose) < 0)
 			goto EXIT;
 	}
